#include <stdio.h>
#include <unistd.h>
#include <rpcsvc/nis.h>
#include <stdlib.h>
#include <pwd.h>
#include <shadow.h>
#include <string.h>
#include <crypt.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>

#define LKUP_FOUND 0x01
#define LKUP_DATAERR 0x02
#define LKUP_ERROR 0x80

extern struct passwd *loguser();

static int lkup(const char *princ,
    const char *domain,
    const char *authname,
    const char *authtype)
{
    nis_result *ptr;
    char buf[256];
    char *cp;
    int retval;

    retval = 0;
    sprintf(buf,"[auth_type=%s,auth_name=%s],cred.org_dir.%s",
		authtype, authname, domain);

    ptr=nis_list(buf,0,NULL,NULL);
    switch(ptr->status)
    {
	case NIS_NOTFOUND:
	    break;
	case NIS_SUCCESS:
	    retval |= LKUP_FOUND ;
	    if (ptr->objects.objects_len != 1)
	    {
		fprintf(stderr,"Unexpected %d records match %s lookup\n",
		    ptr->objects.objects_len,authtype);
		retval |= LKUP_DATAERR ;
	    }
	    if (ptr->objects.objects_val[0].zo_data.zo_type != ENTRY_OBJ)
	    {
		fprintf(stderr,"Unexpected object type %d for %s credential\n",
		    ptr->objects.objects_val[0].zo_data.zo_type, authtype);
		retval |= LKUP_DATAERR ;
		break;
	    }
	    if (strcmp(ENTRY_VAL(ptr->objects.objects_val,0),princ))
	    {
		fprintf(stderr,"Wrong 'cname' for your %s credential\n",
		    authtype);
		retval |= LKUP_DATAERR ;
	    }
		break;
	default:
	    nis_perror(ptr->status,"lkup");
	    retval |= LKUP_ERROR ;
	    break;
    }
    nis_freeresult(ptr);
    return (retval);
}

runcmd(const char *cmd,
    char * const args[])
{
    int pid, status;

    pid = fork();
    if (pid == 0)
    {
	execv(cmd,args);
	fprintf(stderr,"execv(%s) failed\n\n",cmd);
	exit(1);
    }
    if (pid == -1)
    {
	perror("runcmd");
	exit(1);
    }
    while (waitpid(pid, &status, 0) == -1)
    {
	if (errno == EINTR) continue;
	fprintf(stderr,"status is %d\n",status);
	perror ("runcmd");
	exit(1);
    }
    if (WIFEXITED(status)) return (WEXITSTATUS(status));
    fprintf(stderr,"Child process %s exited with status %d\n",
	cmd, status);
    return(status);
}

void main(int argc, char **argv)
{
    char domain[64];
    char localname[12], princ[128],desname[128];
    char *cp;
    int uid = getuid();
    int n;
    struct passwd *pwd;
    struct spwd *sp;
    char *args[10];
    char *cmd = "/usr/bin/nisaddcred";
    int localflag, desflag;

    cp = argc > 1 ? argv[1] : NULL ;
    if (uid == 0)
    {
	fprintf(stderr, "This command is not for root user\n");
	exit(1);
    }
    pwd = loguser(cp);
    if (pwd == NULL)
    {
	fprintf(stderr, "Cannot read password entry for uid=%d\n",
	    uid);
	exit(1);
    }
    if ((cp = nis_local_directory()) == NULL)
    {
	fprintf(stderr, "Cannot get domain name\n");
	exit(1);
    }
    strncpy(domain,cp,sizeof(domain));
    domain[sizeof(domain) - 1] = '\0';
    sprintf(princ,"%s.%s",pwd->pw_name,domain);
    sprintf(localname,"%d",pwd->pw_uid);
    sprintf(desname,"unix.%s@%s",localname,domain);
    desname[strlen(desname) - 1] = '\0';
    sp=getspnam(pwd->pw_name);
    if (sp == NULL)
    {
	fprintf(stderr,"Cannot read shadow data for %s\n",
	    pwd->pw_name);
	exit(1);
    }

    localflag=lkup(princ,domain,localname,"LOCAL");
    if (localflag & LKUP_ERROR) exit(1);
    desflag=lkup(princ,domain,desname,"DES");
    if (desflag & LKUP_ERROR) exit(1);

    if ((localflag | desflag) & LKUP_DATAERR)
    {
	fprintf(stderr,
	    "There is a credentials problem.\n"
	    "Please see system administrator\n\n");
	exit(1);
    }

#ifndef REPLACE_CRED

    if (desflag & LKUP_FOUND)
    {
	fprintf(stderr,
	    "\nYou already have credentials\n\n"
	    "Try to keylogin to validate them\n"
	    "If this fails, see your administrator\n\n");
	exit(1);
    }
#endif /* REPLACE_CRED */

    cp = getpass("Your password: ");
    if (strcmp(crypt(cp,sp->sp_pwdp),sp->sp_pwdp))
    {
	fprintf(stderr,"Sorry\n");
	exit(1);
    }

#ifdef REPLACE_CRED
    if (desflag & LKUP_FOUND)
    {
	char keybuf[HEXKEYBYTES+1];

	if (getsecretkey(desname,keybuf,cp) && keybuf[0])
	{
	    fprintf(stderr,
		"\nYou already have credentials\n\n"
		"Try to keylogin to validate them\n"
		"If this fails, see your administrator\n\n");
	    exit(1);
	}
    }
#endif /* REPLACE_CRED */

    args[0] = cmd;
    args[1] = "-p";
    args[2] = localname ;
    args[3] = "-P" ;
    args[4] = princ;
    args[5] = "local";
    args[6] = NULL;

    if (!(localflag & LKUP_FOUND))
	if (runcmd(cmd,args))
	    exit(1);

    args[2]=desname ;
    args[5] = "-l";
    args[6] = cp ;
    args[7] = "des" ;
    args[8] = NULL ;

    if (runcmd(cmd,args))
	exit(1);
    printf("\nYour credentials have been created.  However it may take 10-15\n"
	     "minutes for the system databases to be updated.  Once that has\n"
	     "been done, you can either use 'keylogin', or you can logout\n"
	     "then relogin using your password, in order to decrypt your\n"
	     "credentials.\n\n");
    exit(0);
}
